--- A module for retrieving sensor and instrument data from the aircraft.

local M = {}

local function toDegreesYaw(r)
    return math.deg(r) % 360
end

local function toDegreesPitch(r)
    local d = math.deg(r)
    if d > 180 then d = d - 360 end
    if d < -180 then d = d + 360 end
    if d > 90 then d = d - 180 end
    if d < -90 then d = d + 180 end
    return d
end

local function toDegreesRoll(r)
    local d = math.deg(r)
    if d > 180 then d = d - 360 end
    if d < -180 then d = d + 360 end
    return d
end

local data = {}

--- Reads aircraft instruments and sensors.
--- @return          table   # aircraft data:
---      altitude            number  altitude (m)
---      trueAirspeed        number  true airspeed (m/s)
---      indicatedAirspeed   number  indicated airspeed (m/s)
---      verticalSpeed       number  vertical speed (m/s):           + for climb,        - for descent
---      yaw                 number  yaw/heading (deg):              + for clockwise,    - for counterclockwise
---      pitch               number  pitch (deg):                    + for up,           - for down
---      roll                number  roll (deg):                     + for clockwise,    - for counterclockwise
function M.getAircraftData()
    local veh = obj
    local airVelocity = veh:getVelocity() - veh:getFlow()
    local roll, pitch, yaw = veh:getRollPitchYaw()
    local densityFactor = veh:getDensityAtNode(0) / 1.225
    local electricsValues = electrics.values

    -- Altitude in m
    data.altitude = veh:getAltitude()
    -- True airspeed in m/s.
    data.trueAirspeed = airVelocity:length()
    -- Indicated airspeed in m/s.
    data.indicatedAirspeed = math.sqrt(airVelocity:squaredLength() * densityFactor)
    -- Vertical speed in m/s; + for climb, - for descent
    data.verticalSpeed = airVelocity.z
    -- Yaw/heading in degrees (increases clockwise). Add 180 deg (pi rad) to align with builtin maps.
    data.yaw = toDegreesYaw(-yaw + math.pi)
    -- Pitch in degrees: + for up, - for down.
    data.pitch = toDegreesPitch(pitch)
    -- Roll in degrees: + for clockwise, - for counterclockwise.
    data.roll = toDegreesRoll(-roll)
    -- Horizontal acceleration.
    -- data.horizontalAcceleration = sensors.gx2 or 0
    -- Aileron deflection [-1, +1]: + to increase bank, - to decrease bank.
    -- data.aileron = 0
    -- Elevator deflection [-1, +1]: + to raise nose, - to lower nose.
    -- data.elevator = 0
    -- Rudder deflection [-1, +1]: + to turn right, - to turn left.
    -- data.rudder = 0
    -- Engine throttle [0, 1].
    -- data.throttle = 0
    return data
end

return M
