--- A module for the main autopilot code.

local M = {}

local modeAlt = require("lua/vehicle/extensions/av331autopilot/modes/alt")
local modeFlc = require("lua/vehicle/extensions/av331autopilot/modes/flc")
local modeHdg = require("lua/vehicle/extensions/av331autopilot/modes/hdg")
local modeSpd = require("lua/vehicle/extensions/av331autopilot/modes/spd")
local modeVs = require("lua/vehicle/extensions/av331autopilot/modes/vs")

local clamp = require("lua/vehicle/extensions/av331autopilot/util").clamp

local modes = {
    modeAlt,
    modeFlc,
    modeVs,
    modeHdg,
    -- SPD must be last, since if A/T is enabled w/o SPD, it is THR.
    modeSpd,
}
local modeCount = #modes

local ctrl = {}

local maxDeflectionAileron = 1.0
local maxDeflectionElevator = 1.0
local maxDeflectionRudder = 1.0
local maxThrottleChange = 1.0

local minDeflectionAileron = -1.0
local minDeflectionElevator = -1.0
local minDeflectionRudder = -1.0
local minThrottleChange = -1.0

--- Applies model-specific parameters.
--- @param param table   model-specific parameters
function M.applyAutopilotParameters(param)
    local paramCopy = {}
    for k, v in pairs(param) do paramCopy[k] = v end

    paramCopy.forceFactor = param.forceFactor or 1
    paramCopy.pitchForceFactor = param.pitchForceFactor or math.min(1, paramCopy.forceFactor ^ 3)
    paramCopy.rollForceFactor = param.rollForceFactor or paramCopy.forceFactor
    paramCopy.throttleForceFactor = param.throttleForceFactor or paramCopy.forceFactor

    for i = 1, modeCount do
        local applyParameters = modes[i].applyParameters
        if applyParameters then
            applyParameters(paramCopy)
        end
    end

    maxDeflectionAileron = paramCopy.maxDeflectionAileron or 1.0
    maxDeflectionElevator = paramCopy.maxDeflectionElevator or 1.0
    maxDeflectionRudder = paramCopy.maxDeflectionRudder or 1.0
    maxThrottleChange = paramCopy.maxThrottleChange or 1.0

    minDeflectionAileron = paramCopy.minDeflectionAileron or -maxDeflectionAileron
    minDeflectionElevator = paramCopy.minDeflectionElevator or -maxDeflectionElevator
    minDeflectionRudder = paramCopy.minDeflectionRudder or -maxDeflectionRudder
    minThrottleChange = paramCopy.minThrottleChange or -maxThrottleChange
end

--- Configures the autopilot.
--- @param autopilot     table   autopilot object
--- @param aircraft      table   aircraft data
function M.configureAutopilot(autopilot, aircraft)
    if not autopilot.enabled then
        autopilot.enabled = {}
    end

    local enabledModes = autopilot.enabled
    for i = 1, modeCount do
        local mode = modes[i]
        local oldEnabled = enabledModes[i] or false
        local newEnabled = (autopilot.on or autopilot.athr) and mode.check(autopilot) or false

        if not oldEnabled and newEnabled and mode.enable then mode.enable(autopilot, aircraft) end
        if oldEnabled and not newEnabled and mode.disable then mode.disable(autopilot, aircraft) end

        enabledModes[i] = newEnabled
    end
end

--- Run the autopilot.
--- @param autopilot     table   autopilot object
--- @param aircraft      table   aircraft data
--- @param delta         number  delta time in seconds
--- @return              table   # Autopilot response
---      aileron         number|nil  aileron deflection within  [-1, +1], -1 = bank left,  +1 = bank right
---      elevator        number|nil  elevator deflection within [-1, +1], -1 = pitch down, +1 = pitch up
---      rudder          number|nil  rudder deflection within   [-1, +1], -1 = yaw left,   +1 = yaw right
---      thrust          number|nil  thrust adjustment within   [-1, +1], -1 = less power, +1 = more power
---      mode            string|nil  new mode (if switching mode)
---      apOff           boolean     if switching off autopilot
---      athrOff         boolean     if switching off autothrottle
---      spdOff          boolean     if switching off SPD mode
---      hdgOff          boolean     if switching off HDG mode
function M.runAutopilot(autopilot, aircraft, delta)
    -- reset
    ctrl.aileron = nil
    ctrl.elevator = nil
    ctrl.rudder = nil
    ctrl.throttle = nil
    ctrl.mode = nil
    ctrl.apOff = false
    ctrl.athrOff = false
    ctrl.spdOff = false
    ctrl.hdgOff = false

    if autopilot.enabled then
        for i = 1, modeCount do
            if autopilot.enabled[i] then
                modes[i].run(autopilot, aircraft, delta, ctrl)
            end
        end
    end

    if not autopilot.on then
        ctrl.aileron = nil
        ctrl.elevator = nil
        ctrl.rudder = nil
    else
        if ctrl.aileron then ctrl.aileron = clamp(ctrl.aileron, minDeflectionAileron, maxDeflectionAileron) end
        if ctrl.elevator then ctrl.elevator = clamp(ctrl.elevator, minDeflectionElevator, maxDeflectionElevator) end
        if ctrl.rudder then ctrl.rudder = clamp(ctrl.rudder, minDeflectionRudder, maxDeflectionRudder) end
    end
    if not autopilot.athr then
        ctrl.throttle = nil
    else
        if ctrl.throttle then ctrl.throttle = clamp(ctrl.throttle, minThrottleChange, maxThrottleChange) end
    end

    return ctrl
end

return M
