--- A module for issuing controls to an aircraft.

local M = {}

local util = require("lua/vehicle/extensions/av331autopilot/util")
local clamp = util.clamp

local function submitInputEvent(name, val)
    input.event(name, val, "FILTER_AI", nil, nil, nil, "autopilot")
end

--- Issues controls to the aircraft.
--- @param ctrl          table       controls to apply
--- @param ctrl.aileron  number|nil  aileron deflection within  [-1, +1], -1 = bank left,  +1 = bank right
--- @param ctrl.elevator number|nil  elevator deflection within [-1, +1], -1 = pitch down, +1 = pitch up
--- @param ctrl.rudder   number|nil  rudder deflection within   [-1, +1], -1 = yaw left,   +1 = yaw right
--- @param ctrl.thrust   number|nil  thrust adjustment within   [-1, +1], -1 = less power, +1 = more power
local function issueControls_generic(ctrl)
    if ctrl.aileron  then submitInputEvent("aileron",   ctrl.aileron)  end
    if ctrl.elevator then submitInputEvent("elevator", -ctrl.elevator) end
    if ctrl.rudder   then submitInputEvent("rudder",    ctrl.rudder)   end
    if ctrl.throttle then submitInputEvent("throttle",  ctrl.throttle) end
end

local function issueControls_impl_B25Mitchell(ctrl)
    if ctrl.aileron  then submitInputEvent("aileron",   ctrl.aileron)  end
    if ctrl.elevator then submitInputEvent("elevator", -ctrl.elevator) end
    if ctrl.rudder   then submitInputEvent("rudder",    ctrl.rudder)   end
    if ctrl.throttle then
        electrics.values.throttleB25Increase = math.max(0, ctrl.throttle)
        electrics.values.throttleB25Decrease = math.max(0, -ctrl.throttle)
    end
end

local function issueControls_impl_me262(ctrl)
    if ctrl.aileron  then submitInputEvent("aileron",   ctrl.aileron)  end
    if ctrl.elevator then submitInputEvent("elevator",  ctrl.elevator) end
    if ctrl.rudder   then submitInputEvent("rudder",    ctrl.rudder)   end
    if ctrl.throttle then
        electrics.values.jet_throttle_change = ctrl.throttle
    end
end

local mtngoat_cachedControls = {}
local mtngoat_liveControls = {}

local function updateControlImpl(old, live, target, vmin, vmax)
    if live == nil then return target end
    if old == nil then old = live end
    return clamp(target + live - old, vmin, vmax)
end

local function updateControl(cache, live, input, name, vmin, vmax)
    local result
    local target = input[name]
    if target then
        result = updateControlImpl(cache[name], live[name], target, vmin, vmax)
        cache[name] = result
        return result
    else
        cache[name] = live[name]
        return nil
    end
end

local function issueControls_impl_mtngoat(ctrl)
    local cachedControls = mtngoat_cachedControls
    local liveControls = mtngoat_liveControls
    local electricsValues = electrics.values
    local result

    liveControls.rudder = electricsValues.yaw or 0
    liveControls.elevator = -(electricsValues.pitch or 0)
    liveControls.aileron = electricsValues.roll or 0
    liveControls.throttle = (electricsValues.throttleup or 0) - (electricsValues.throttledown or 0)

    result = updateControl(cachedControls, liveControls, ctrl, "rudder", -1, 1)
    if result ~= nil then electricsValues.yaw = result end

    result = updateControl(cachedControls, liveControls, ctrl, "elevator", -1, 1)
    if result ~= nil then electricsValues.pitch = -result end

    result = updateControl(cachedControls, liveControls, ctrl, "aileron", -1, 1)
    if result ~= nil then electricsValues.roll = result end

    result = updateControl(cachedControls, liveControls, ctrl, "throttle", -1, 1)
    if result ~= nil then
        electricsValues.throttleup = math.max(0, result)
        electricsValues.throttledown = math.max(0, -result)
    end
end

local issueControls_impl = {
    ["mtngoat"] = issueControls_impl_mtngoat,
    ["B25Mitchell"] = issueControls_impl_B25Mitchell,
    ["me262"] = issueControls_impl_me262,
}

function M.onReset()
    mtngoat_cachedControls = {}
end

--- Returns the aircraft control issue function for a particular model.
--- @param      model   string|nil          model name or nil for generic
--- @return             fun(ctrl: table)    # aircraft control function
function M.issueControlsFactory(model)
    return issueControls_impl[model] or issueControls_generic
end

return M
