--- Autopilot ALT (altitude hold) mode

local M = {}

local pid = require("lua/vehicle/extensions/av331autopilot/pid")
local pidStep = pid.autopilotPidStep

local clamp = require("lua/vehicle/extensions/av331autopilot/util").clamp

-- Elevator force factor.
local pitchForceFactor = 1.0
-- Throttle force factor.
local throttleForceFactor = 1.0

-- Altitude -> vertical/speed multiplier, e.g. to gain +10 m, target V/S is 0.4/s * +10m = +4m/s
local altVsMultiplier = 0.4

-- Paramters for adjusting power/thrust for ALT mode
local altPowerAltMultiplier = 2     -- Altitude difference weight
local altPowerVsMultiplier = 1.5    -- V/S weight
local altPowerFbMultiplier = 10     -- feedback weight
local altPowerMinAltDiff = -10      -- If alt diff < -10 m, then treated as -10 m
local altPowerMinVsDiff = -20       -- If V/S < -20 m/s, then treated as -20 m/s
local altPowerMaxVsDiff = 40        -- If V/S > +40 m/s, then treated as +40 m/s
local altPowerBuf = 5               -- once integrated value goes < -5, start adding power
local altPowerMaxAdd = 25           -- integrated value minimum value is -25 = -(5 + 20)
local altPowerExp = 1               -- exponential curve for adding power
local altPowerMul = 5               -- constant factor for adding power

local pitchTargetPidParams = {
    Kp = 0.36, Ki = 1.2, Kd = 0.0,
    cMin = -30, cMax = 30, eMin = -30, eMax = 30
}

local elevatorPidParams = {
    Kp = 0.1, Ki = 0.02, Kd = 0.0,
    cMin = -1, cMax = 1, eMin = -1, eMax = 1
}

--- Applies model-specific parameters.
--- @param param table   model-specific parameters
function M.applyParameters(param)
    pitchForceFactor = param.pitchForceFactor or pitchForceFactor
    throttleForceFactor = param.throttleForceFactor or throttleForceFactor
    altVsMultiplier = param.altVsMultiplier or altVsMultiplier
    altPowerAltMultiplier = param.altPowerAltMultiplier or altPowerAltMultiplier
    altPowerVsMultiplier = param.altPowerVsMultiplier or altPowerVsMultiplier
    altPowerMinAltDiff = param.altPowerMinAltDiff or altPowerMinAltDiff
    altPowerMinVsDiff = param.altPowerMinVsDiff or altPowerMinVsDiff
    altPowerMaxVsDiff = param.altPowerMaxVsDiff or altPowerMaxVsDiff
    altPowerBuf = param.altPowerBuf or altPowerBuf
    altPowerMaxAdd = param.altPowerMaxAdd or altPowerMaxAdd
    altPowerExp = param.altPowerExp or altPowerExp
    altPowerMul = param.altPowerMul or altPowerMul
    
    pitchTargetPidParams.cMin = param.pitchMin or pitchTargetPidParams.cMin
    pitchTargetPidParams.cMax = param.pitchMax or pitchTargetPidParams.cMax
    pitchTargetPidParams.eMin = param.pitchMin or pitchTargetPidParams.eMin
    pitchTargetPidParams.eMax = param.pitchMax or pitchTargetPidParams.eMax

    local maxDeflectionElevator = param.maxDeflectionElevator or 1.0
    local minDeflectionElevator = param.minDeflectionElevator or -maxDeflectionElevator
    elevatorPidParams.cMin = minDeflectionElevator
    elevatorPidParams.eMin = minDeflectionElevator
    elevatorPidParams.cMax = maxDeflectionElevator
    elevatorPidParams.eMax = maxDeflectionElevator

    pid.applyPidParameters(pitchTargetPidParams, param.altPitchTargetPid)
    pid.applyPidParameters(elevatorPidParams, param.altElevatorPid)
end

local altPowerDiv = altPowerMul / (altPowerMaxAdd ^ (altPowerExp - 1))
local lastThrottle = 0

--- Checks whether the autopilot should run ALT mode.
--- @param ap    table   autopilot object
--- @return      boolean # whether ALT is enabled
function M.check(ap)
    return ap.mode == "alt"
end

--- Called when autopilot ALT (altitude hold) mode is engaged.
--- @param ap    table   autopilot object
--- @param ac    table   aircraft sensor/instrument data
function M.enable(ap, ac)
    -- set initial pitch to prevent a swing when mode engaged.
    pid.resetPid(ap, "altPitchTarget", ac.pitch)
    ap.morePowerInt = 0
    lastThrottle = 0
end

--- Run the autopilot ALT (altitude hold) mode.
--- @param ap    table   autopilot object
--- @param ac    table   aircraft sensor/instrument data
--- @param dt    number  delta time in seconds
--- @param ctrl  table   controls to apply in the format used by issueAircraftControls; fields may be missing
function M.run(ap, ac, dt, ctrl)
    if ap.alt == nil then
        ctrl.mode = ""
        return
    end

	local err = (ap.alt - ac.altitude) * altVsMultiplier - ac.verticalSpeed
    local tmp = pidStep(ap, "altPitchTarget", pitchTargetPidParams, dt, pitchForceFactor * err)
	local out = pidStep(ap, "altPitch", elevatorPidParams, dt, tmp - ac.pitch)
    ctrl.elevator = (ctrl.elevator or 0) + out

    local normAltDiff = math.max(altPowerMinAltDiff, ac.altitude - ap.alt)
    local normVs = clamp(ac.verticalSpeed, altPowerMinVsDiff, altPowerMaxVsDiff)
    if normAltDiff > 0 then
        normVs = normVs + normAltDiff
    elseif normVs > 0 then
        normAltDiff = normAltDiff + normVs
    end

    ap.morePowerInt = clamp(ap.morePowerInt +
            dt * (altPowerAltMultiplier * normAltDiff
                + altPowerVsMultiplier * normVs
                + altPowerFbMultiplier * lastThrottle), -altPowerBuf - altPowerMaxAdd, 0)
    if ap.morePowerInt < -altPowerBuf then
        local powerScale = -ap.morePowerInt - altPowerBuf
        local newThrottle = dt * throttleForceFactor * powerScale * powerScale ^ altPowerExp * altPowerDiv
        ctrl.throttle = (ctrl.throttle or 0) + newThrottle
        lastThrottle = newThrottle
    else
        lastThrottle = 0
    end
end

return M
