--- Autopilot FLC (flight level change) mode

local M = {}

local pid = require("lua/vehicle/extensions/av331autopilot/pid")
local pidStep = pid.autopilotPidStep

--- Checks whether the autopilot should run FLC mode.
--- @param ap    table   autopilot object
--- @return      boolean # whether FLC is enabled
function M.check(ap)
    return ap.mode == "flc"
end

-- Elevator force factor.
local pitchForceFactor = 1.0
local flcDampenWithSpd = 0.1

local pitchTargetPidParams = {
    Kp = 0.35, Ki = 0.55, Kd = 0.01,
    cMin = -30, cMax = 30, eMin = -30, eMax = 30
}

local elevatorPidParams = {
    Kp = 0.1, Ki = 0.02, Kd = 0.005,
    cMin = -1, cMax = 1, eMin = -1, eMax = 1
}

--- Applies model-specific parameters.
--- @param param table   model-specific parameters
function M.applyParameters(param)
    pitchForceFactor = param.pitchForceFactor or pitchForceFactor
    flcDampenWithSpd = param.flcDampenWithSpd or flcDampenWithSpd
    pitchTargetPidParams.cMin = param.pitchMin or pitchTargetPidParams.cMin
    pitchTargetPidParams.cMax = param.pitchMax or pitchTargetPidParams.cMax
    pitchTargetPidParams.eMin = param.pitchMin or pitchTargetPidParams.eMin
    pitchTargetPidParams.eMax = param.pitchMax or pitchTargetPidParams.eMax

    local maxDeflectionElevator = param.maxDeflectionElevator or 1.0
    local minDeflectionElevator = param.minDeflectionElevator or -maxDeflectionElevator
    elevatorPidParams.cMin = minDeflectionElevator
    elevatorPidParams.eMin = minDeflectionElevator
    elevatorPidParams.cMax = maxDeflectionElevator
    elevatorPidParams.eMax = maxDeflectionElevator

    pid.applyPidParameters(pitchTargetPidParams, param.flcPitchTargetPid)
    pid.applyPidParameters(elevatorPidParams, param.flcElevatorPid)
end

--- Called when autopilot ALT (altitude hold) mode is engaged.
--- @param ap    table   autopilot object
--- @param ac    table   aircraft sensor/instrument data
function M.enable(ap, ac)
    -- set initial pitch to prevent a swing when mode engaged.
    pid.resetPid(ap, "flcPitchTarget", ac.pitch)
end

--- Run the autopilot FLC (flight level change) mode.
--- @param ap    table   autopilot object
--- @param ac    table   aircraft sensor/instrument data
--- @param dt    number  delta time in seconds
--- @param ctrl  table   controls to apply in the format used by issueAircraftControls; fields may be missing
function M.run(ap, ac, dt, ctrl)
    if ap.spd == nil then
        ctrl.mode = ""
        return
    end

    local err = ac.indicatedAirspeed - ap.spd
    local dampen = (ap.athr and ap.spdOn and flcDampenWithSpd) or 1
    local pitchTarget = pidStep(ap, "flcPitchTarget", pitchTargetPidParams, dt, pitchForceFactor * err * dampen)
	local out = pidStep(ap, "flcPitch", elevatorPidParams, dt, pitchTarget - ac.pitch)
    ctrl.elevator = (ctrl.elevator or 0) + out

    if ap.alt == nil then return end
    -- If reaching target altitude within three seconds, switch to ALT.
    local altDiff = ap.alt - ac.altitude
    if math.abs(altDiff / math.max(ac.verticalSpeed, 0.01)) < 3 then
        ctrl.mode = "alt"
    end
end

return M
