--- Autopilot HDG (heading) mode

local M = {}

local pid = require("lua/vehicle/extensions/av331autopilot/pid")
local pidStep = pid.autopilotPidStep

--- Checks whether the autopilot should run HDG mode.
--- @param ap    table   autopilot object
--- @return      boolean # whether HDG is enabled
function M.check(ap)
    return ap.hdgOn
end

local function yawDifference(a, b)
    local d = (a - b) % 360
    if d > 180 then d = d - 360 end
    return d
end

-- Aileron force factor.
local rollForceFactor = 1.0

-- Auto-coordination (0 to disable); applies rudder input based on yaw input.
local autocoordinationFactor = 0.5

local rollTargetPidParams = {
    Kp = 1, Ki = 0.4, Kd = 0.01,
    cMin = -30, cMax = 30, eMin = -5, eMax = 5
}

local rollPidParams = {
    Kp = 0.225, Ki = 0.002, Kd = 0.001,
    cMin = -1, cMax = 1, eMin = -1, eMax = 1
}

--- Applies model-specific parameters.
--- @param param table   model-specific parameters
function M.applyParameters(param)
    rollForceFactor = param.rollForceFactor or rollForceFactor
    rollTargetPidParams.cMin = param.rollMin or rollTargetPidParams.cMin
    rollTargetPidParams.cMax = param.rollMax or rollTargetPidParams.cMax
    rollTargetPidParams.eMin = param.rollMin or rollTargetPidParams.eMin
    rollTargetPidParams.eMax = param.rollMax or rollTargetPidParams.eMax
    autocoordinationFactor = param.autocoordinationFactor or autocoordinationFactor

    local maxDeflectionAileron = param.maxDeflectionAileron or 1.0
    local minDeflectionAileron = param.minDeflectionAileron or -maxDeflectionAileron
    rollPidParams.eMin = minDeflectionAileron
    rollPidParams.cMin = minDeflectionAileron
    rollPidParams.eMax = maxDeflectionAileron
    rollPidParams.cMax = maxDeflectionAileron

    pid.applyPidParameters(rollTargetPidParams, param.hdgRollTargetPid)
    pid.applyPidParameters(rollPidParams, param.hdgRollPid)
end

--- Called when autopilot ALT (altitude hold) mode is engaged.
--- @param ap    table   autopilot object
--- @param ac    table   aircraft sensor/instrument data
function M.enable(ap, ac)
    -- set initial roll to prevent a swing when mode engaged.
    pid.resetPid(ap, "hdgRollTarget", ac.roll)
end

--- Run the autopilot HDG (heading) mode.
--- @param ap    table   autopilot object
--- @param ac    table   aircraft sensor/instrument data
--- @param dt    number  delta time in seconds
--- @param ctrl  table   controls to apply in the format used by issueAircraftControls; fields may be missing
function M.run(ap, ac, dt, ctrl)
    if ap.hdg == nil then
        ctrl.hdgOff = true
        return
    end

    local tmp = pidStep(ap, "hdgRollTarget", rollTargetPidParams, dt, rollForceFactor * yawDifference(ap.hdg, ac.yaw))
    local out = pidStep(ap, "hdgRoll", rollPidParams, dt, tmp - ac.roll)
    ctrl.aileron = (ctrl.aileron or 0) + out

    -- try to apply rudder with autocoordination
    if autocoordinationFactor ~= 0 then
        ctrl.rudder = (ctrl.rudder or 0) + (out * autocoordinationFactor)
    end
end

return M
