--- Autopilot SPD (airspeed) mode

local M = {}

local pid = require("lua/vehicle/extensions/av331autopilot/pid")
local pidStep = pid.autopilotPidStep

--- Checks whether the autopilot should run SPD mode.
--- @param ap    table   autopilot object
--- @return      boolean # whether SPD is enabled
function M.check(ap)
    return ap.spdOn
end

-- Throttle force factor.
local throttleForceFactor = 1.0

local pidParams = {
    Kp = 1.0, Ki = 0, Kd = 1.0,
    cMin = -1, cMax = 1, eMin = -1, eMax = 1
}

--- Applies model-specific parameters.
--- @param param table   model-specific parameters
function M.applyParameters(param)
    throttleForceFactor = param.throttleForceFactor or throttleForceFactor
    local maxThrottleChange = param.maxThrottleChange or 1.0
    local minThrottleChange = param.minThrottleChange or -maxThrottleChange
    pidParams.cMin = minThrottleChange
    pidParams.eMin = minThrottleChange
    pidParams.cMax = maxThrottleChange
    pidParams.eMax = maxThrottleChange

    pid.applyPidParameters(pidParams, param.spdSpeedPid)
end

--- Run the autopilot SPD (airspeed) mode.
--- @param ap    table   autopilot object
--- @param ac    table   aircraft sensor/instrument data
--- @param dt    number  delta time in seconds
--- @param ctrl  table   controls to apply in the format used by issueAircraftControls; fields may be missing
function M.run(ap, ac, dt, ctrl)
    if ap.spd == nil then
        ctrl.spdOff = true
        return
    end

    local err = ap.spd - ac.indicatedAirspeed
    local out = pidStep(ap, "spdThrottle", pidParams, dt, throttleForceFactor * err)
    -- if not on SPD, A/T is on THR mode and takes speed from other modes
    ctrl.throttle = out
end

return M
