--- Autopilot V/S (vertical speed) mode

local M = {}

local pid = require("lua/vehicle/extensions/av331autopilot/pid")
local pidStep = pid.autopilotPidStep

--- Checks whether the autopilot should run V/S mode.
--- @param ap    table   autopilot object
--- @return      boolean # whether V/S is enabled
function M.check(ap)
    return ap.mode == "vs"
end

-- Elevator forcef actor.
local pitchForceFactor = 1.0

local pitchTargetPidParams = {
    Kp = 0.36, Ki = 1.2, Kd = 0.0,
    cMin = -30, cMax = 30, eMin = -30, eMax = 30
}

local elevatorPidParams = {
    Kp = 0.1, Ki = 0.02, Kd = 0.005,
    cMin = -1, cMax = 1, eMin = -1, eMax = 1
}

--- Applies model-specific parameters.
--- @param param table   model-specific parameters
function M.applyParameters(param)
    pitchForceFactor = param.pitchForceFactor or pitchForceFactor
    pitchTargetPidParams.cMin = param.pitchMin or pitchTargetPidParams.cMin
    pitchTargetPidParams.cMax = param.pitchMax or pitchTargetPidParams.cMax
    pitchTargetPidParams.eMin = param.pitchMin or pitchTargetPidParams.eMin
    pitchTargetPidParams.eMax = param.pitchMax or pitchTargetPidParams.eMax

    local maxDeflectionElevator = param.maxDeflectionElevator or 1.0
    local minDeflectionElevator = param.minDeflectionElevator or -maxDeflectionElevator
    elevatorPidParams.cMin = minDeflectionElevator
    elevatorPidParams.eMin = minDeflectionElevator
    elevatorPidParams.cMax = maxDeflectionElevator
    elevatorPidParams.eMax = maxDeflectionElevator

    pid.applyPidParameters(pitchTargetPidParams, param.vsPitchTargetPid)
    pid.applyPidParameters(elevatorPidParams, param.vsElevatorPid)
end

--- Called when autopilot ALT (altitude hold) mode is engaged.
--- @param ap    table   autopilot object
--- @param ac    table   aircraft sensor/instrument data
function M.enable(ap, ac)
    -- set initial pitch to prevent a swing when mode engaged.
    pid.resetPid(ap, "vsPitchTarget", ac.pitch)
end

--- Run the autopilot V/S (vertical speed) mode.
--- @param ap    table   autopilot object
--- @param ac    table   aircraft sensor/instrument data
--- @param dt    number  delta time in seconds
--- @param ctrl  table   controls to apply in the format used by issueAircraftControls; fields may be missing
function M.run(ap, ac, dt, ctrl)
    if ap.vs == nil then
        ctrl.mode = ""
        return
    end

    local err = ap.vs - ac.verticalSpeed
    local pitchTarget = pidStep(ap, "vsPitchTarget", pitchTargetPidParams, dt, pitchForceFactor * err)
	local out = pidStep(ap, "vsPitch", elevatorPidParams, dt, pitchTarget - ac.pitch)
    ctrl.elevator = (ctrl.elevator or 0) + out

    if ap.alt == nil then return end
    -- If reaching target altitude within three seconds, switch to ALT.
    local altDiff = ap.alt - ac.altitude
    if math.abs(altDiff / math.max(ac.verticalSpeed, 0.01)) < 3 then
        ctrl.mode = "alt"
    end
end

return M
