--- PID (proportional-integral-derivative) controller.

local M = {}

local util = require("lua/vehicle/extensions/av331autopilot/util")
local clamp = util.clamp

-- A single PID controller step for PID controller p, parameters c, delta time dt (seconds) and specified error err
local function pidStep(p, c, dt, err)
    local int, prev = p[1], p[2]
    local der

    int = clamp(int + c.Ki * dt * 0.5 * (err + prev), c.eMin, c.eMax)
    if c.Kd > 0 and dt > 1e-10 then
        der = (err - prev) / dt
    else
        der = 0
    end

    p[1], p[2] = int, err
    local cmd = c.Kp * err + int + c.Kd * der
    -- log("d", string.format("av331-pid-%s", p.name), string.format("e=%+1.6f P=%+1.6f I=%+1.6f D=%+1.6f P+I+D=%+1.6f", err, c.Kp * err, int, c.Kd * der, cmd))
    return clamp(cmd, c.cMin, c.cMax)
end

--- Runs a single PID controller step for an autopilot PID by name.
--- @param autopilot     table      autopilot object
--- @param pidName       string     PID controller name
--- @param params        table      PID controller parameters
--- @param params.Kp     number     proportional term coefficient (P)
--- @param params.Ki     number     integral term coefficient (I)
--- @param params.Kd     number     derivative term coefficient (D)
--- @param params.eMin   number     minimum integrated error (saturated)
--- @param params.eMax   number     maximum integrated error (saturated)
--- @param params.cMin   number     minimum command (saturated)
--- @param params.cMax   number     maximum command (saturated)
--- @param dt            number     delta time (seconds)
--- @param err           number     measured error, e.g. setpoint - measurement
--- @return              number     # command from PID controller within [cMin, cMax]
function M.autopilotPidStep(autopilot, pidName, params, dt, err)
    local pids = autopilot.pid
    if not pids then
        pids = {}
        autopilot.pid = pids
    end
    if not pids[pidName] then
        -- integral, previous error
        pids[pidName] = { 0, 0, ["name"] = pidName }
    end
    return pidStep(pids[pidName], params, dt, err)
end

--- Resets a PID controller for an autopilot.
--- @param autopilot     table      autopilot object
--- @param pidName       string     PID controller name
--- @param integ         number|nil default integrator value; 0 if not included
function M.resetPid(autopilot, pidName, integ)
    if not autopilot.pid then
        autopilot.pid = {}
    end
    autopilot.pid[pidName] = { integ or 0, 0, ["name"] = pidName }
end

--- Applies PID controller parameters.
--- @param target        table      PID parameters to modify
--- @param source        table      PID parameter source
function M.applyPidParameters(target, source)
    if source then
        for k, v in pairs(target) do
            target[k] = source[k] or v
        end
    end
end

--- Calibrates PID with Ziegler-Nichols.
--- @param ku            number     K_u (ultimate gain)
--- @param tu            number     T_u (oscillation period)
--- @param overshoot     string     "classic", "pessen", "some" or "none"
--- @param params        table      remaining PID controller parameters
--- @param params.eMin   number     minimum integrated error (saturated)
--- @param params.eMax   number     maximum integrated error (saturated)
--- @param params.cMin   number     minimum command (saturated)
--- @param params.cMax   number     maximum command (saturated)
--- @return              table      # parameters with Kp, Ki, Kd populated
function M.pidTunedZN(ku, tu, overshoot, params)
    local kp, ti, td
    if overshoot == "pessen" then
        kp, ti, td = 0.7  * ku, 0.4 * tu, 0.15  * tu
    elseif overshoot == "some" then
        kp, ti, td = 0.33 * ku, 0.5 * tu, 0.33  * tu
    elseif overshoot == "none" then
        kp, ti, td = 0.2  * ku, 0.5 * tu, 0.33  * tu
    else -- overshoot == "classic"
        kp, ti, td = 0.6  * ku, 0.5 * tu, 0.125 * tu
    end

    params.Kp = kp
    params.Ki = kp / ti
    params.Kd = kp * td
    return params
end

return M
