--- Aircraft autopilot interface for BeamNG.drive vehicle engine and UI

local M = {}

local aircraftDataModule = require("lua/vehicle/extensions/av331autopilot/aircraftData")
local newAutopilot = require("lua/vehicle/extensions/av331autopilot/autopilotData").newAutopilot
local autopilotMain = require("lua/vehicle/extensions/av331autopilot/autopilotMain")
local issueControlsModule = require("lua/vehicle/extensions/av331autopilot/issueControls")
local modelDetect = require("lua/vehicle/extensions/av331autopilot/modelDetect")
local parameters = require("lua/vehicle/extensions/av331autopilot/parameters")

local runAutopilot = autopilotMain.runAutopilot

local autopilot = nil
local model = nil

local getAircraftDataRaw = aircraftDataModule.getAircraftData

--- Issues controls to the aircraft.
--- @param ctrl          table       controls to apply
--- @param ctrl.aileron  number|nil  aileron deflection within  [-1, +1], -1 = bank left,  +1 = bank right
--- @param ctrl.elevator number|nil  elevator deflection within [-1, +1], -1 = pitch down, +1 = pitch up
--- @param ctrl.rudder   number|nil  rudder deflection within   [-1, +1], -1 = yaw left,   +1 = yaw right
--- @param ctrl.thrust   number|nil  thrust adjustment within   [-1, +1], -1 = less power, +1 = more power
local issueControls = nil

local getAircraftDataHook = nil

-- whether to send autopilot data through stream
local refreshAutopilotInStream = true
local forceSendAutopilot = true

---------------------
-- BASIC FUNCTIONS --
---------------------

--- Reads aircraft instruments and sensors.
--- @return          table   # aircraft data:
---      altitude            number  altitude (m)
---      trueAirspeed        number  true airspeed (m/s)
---      indicatedAirspeed   number  indicated airspeed (m/s)
---      verticalSpeed       number  vertical speed (m/s):           - for descent           + for climb
---      yaw                 number  yaw/heading (deg):              - for counterclockwise  + for clockwise
---      pitch               number  pitch (deg):                    - for down              + for up
---      roll                number  roll (deg):                     - for counterclockwise  + for clockwise
local function getAircraftData()
    local data = getAircraftDataRaw()
    if getAircraftDataHook then
        getAircraftDataHook(data)
    end
    return data
end

-- Announce autopilot change to GELUA, which may announce it to the UI.
local function onAutopilotChanged()
    if not autopilot then return end
    refreshAutopilotInStream = true
end

--- Computes default roughness based on mass.
local function computeDefaultRoughness()
    local fullRoughnessAtWeight = 616
    local partialRoughnessAtWeight = 14196
    local partialRoughness = 0.5

    local targetValueAfterExponent = 1 / partialRoughness
    local exponent = math.log(targetValueAfterExponent) / math.log(partialRoughnessAtWeight / fullRoughnessAtWeight)

    local totalWeight = obj:calcBeamStats().total_weight
    local weightWithExponent = (totalWeight / fullRoughnessAtWeight) ^ exponent
    return 1 / weightWithExponent
end

local setupExecuted = false
local canStillCustomSetup = true
--- Initializes the vehicle autopilot if it has not been initialized yet.
--- This is only ever called when the user interacts with the autopilot, since
--- it is entirely off/disabled by default.
function M.setup()
    if setupExecuted then return end
    setupExecuted = true

    parameters.genericParameters.roughness = computeDefaultRoughness()
    autopilotMain.applyAutopilotParameters(parameters.genericParameters)
    model = nil

    -- allow vehicles to hook model detection and provide their own overrides
    extensions.hook("onAv331AutopilotSetup")
    canStillCustomSetup = false

    if not model then
        model = modelDetect.detectAircraftModel()
        if model then
            local modelParameters = parameters.modelParameters[model]
            if modelParameters then
                autopilotMain.applyAutopilotParameters(modelParameters)
            end
        end
    end

    if not issueControls then
        issueControls = issueControlsModule.issueControlsFactory(model)
    end

    if not getAircraftDataHook then
        getAircraftData = getAircraftDataRaw
    end

    autopilot = newAutopilot(model)
    autopilotMain.configureAutopilot(autopilot, getAircraftData())

    log("d", "av331autopilot", string.format("Detected aircraft model is %s", model and "'"..model.."'" or "<generic>"))
    refreshAutopilotInStream = true
end

local hasCalledSetup = false
--- Initializes model-specific details; meant to be called by a vehicle's onAv331AutopilotSetup callback.
--- @param modelName                    string              model name
--- @param setup                        table               custom setup.
--- @param setup.issueControls          fun(table)          a function that issues controls to the aircraft
---     issueControls' parameter is a table containing the following values (all are optional and may be nil if the autopilot does not input anything):
---         aileron                     number      [minDeflectionAileron, maxDeflectionAileron]: 0 for center ailerons, - to bank left, + to bank right
---         elevator                    number      [minDeflectionElevator, maxDeflectionElevator]: 0 to center elevator, - to pitch nose down, + to pitch nose up
---         rudder                      number      [minDeflectionRudder, maxDeflectionRudder]: 0 to center rudder, - to yaw left, + to yaw right
---         throttle                    number      [minThrottleChange, maxThrottleChange]: 0 to not adjust power, - to reduce power, + to add power
--- @param setup.getAircraftDataHook    fun(table)          a function that obtains aircraft data. you rarely need this.
---     getAircraftDataHook's parameter is a table containing at least the following values, which you can modify if needed:
---         altitude                    number      'barometric' alt in meters. from the Z coordinate; does not care about sea level
---         trueAirspeed                number      true airspeed in m/s
---         indicatedAirspeed           number      indicated airspeed in m/s (true airspeed adjusted based on air density)
---         verticalSpeed               number      vertical speed in m/s; + for climb, - for descent
---         yaw                         number      yaw/heading in degrees. 0 degrees is as BeamNG defines it; angle increases clockwise.
---         pitch                       number      pitch in degrees. 0 degrees for nose level, + for nose up, - for nose down.
---         roll                        number      roll/bank in degrees. 0 degrees for wings level, + for bank right, - for bank left.
--- @param setup.parameters             table               various other parameters, all values are optional unless otherwise specified.
---   BASIC PARAMETERS
---     forceFactor                     number      scales the rate/amount of pitch/roll input           (default: computed based on aircraft mass)
---     pitchForceFactor                number      scales the rate/amount of pitch input                (default: forceFactor^3 or 1 whichever is less)
---     rollForceFactor                 number      scales the rate/amount of roll input                 (default: forceFactor)
---     throttleForceFactor             number      scales the rate/amount of throttle input             (default: forceFactor)
---     autocoordinationFactor          number      auto-coordination: rudder input relative to aileron  (default: 0.5)
---     maxDeflectionAileron            number      maximum aileron deflection when banking right        (default: +1)
---     maxDeflectionElevator           number      maximum elevator deflection when pitching nose up    (default: +1)
---     maxDeflectionRudder             number      maximum rudder deflection when turning right         (default: +1)
---     maxThrottleChange               number      maximum throttle change when adding power            (default: +1)
---     minDeflectionAileron            number      minimum aileron deflection when banking left         (default: negative of maxDeflectionAileron)
---     minDeflectionElevator           number      minimum elevator deflection when pitching nose down  (default: negative of maxDeflectionElevator)
---     minDeflectionRudder             number      minimum rudder deflection when turning left          (default: negative of maxDeflectionRudder)
---     minThrottleChange               number      minimum throttle change when removing power          (default: negative of maxThrottleChange)
--- 
---   ALT MODE PARAMETERS
---     altVsMultiplier                 number      ALT mode: altitude -> vertical speed multiplier, e.g. to gain +10 m, target V/S is 0.4/s * +10m = +4m/s         (default: 0.4)
---     altPowerAltMultiplier           number      ALT/THR mode (add power in ALT if insufficient): Altitude difference weight                                     (default: 2)
---     altPowerVsMultiplier            number      ALT/THR mode (add power in ALT if insufficient): V/S weight                                                     (default: 1.5)
---     altPowerFbMultiplier            number      ALT/THR mode (add power in ALT if insufficient): feedback weight                                                (default: 10)
---     altPowerMinAltDiff              number      ALT/THR mode (add power in ALT if insufficient): If alt diff < -10 m, then treated as -10 m                     (default: -10)
---     altPowerMinVsDiff               number      ALT/THR mode (add power in ALT if insufficient): If V/S < -20 m/s, then treated as -20 m/s                      (default: -20)
---     altPowerMaxVsDiff               number      ALT/THR mode (add power in ALT if insufficient): If V/S > +40 m/s, then treated as +40 m/s                      (default: 40)
---     altPowerBuf                     number      ALT/THR mode (add power in ALT if insufficient): once integrated value goes < -5, start adding power            (default: 5)
---     altPowerMaxAdd                  number      ALT/THR mode (add power in ALT if insufficient): integrated value minimum value is -25 = -(5 + 20)              (default: 20)
---     altPowerExp                     number      ALT/THR mode (add power in ALT if insufficient): exponential curve for adding power                             (default: 1)
---     altPowerMul                     number      ALT/THR mode (add power in ALT if insufficient): constant factor for adding power                               (default: 5)
--- 
---   FLC MODE PARAMETERS *
---     flcDampenWithSpd                number      FLC dampen factor when SPD enabled to prevent oscillation (default: 0.1)
--- 
---   PID TUNING PARAMETERS
---     altPitchTargetPid               table       PID parameters for ALT mode pitch angle target
---     altElevatorPid                  table       PID parameters for elevator adjustments in ALT mode
---     flcPitchTargetPid               table       PID parameters for FLC mode pitch angle target
---     flcElevatorPid                  table       PID parameters for elevator adjustments in FLC mode
---     hdgRollTargetPid                table       PID parameters for HDG mode roll angle target
---     hdgRollPid                      table       PID parameters for aileron adjustments in HDG mode
---     spdSpeedPid                     table       PID parameters for throttle adjustments in SPD mode
---     vsPitchTargetPid                table       PID parameters for V/S mode pitch angle target
---     vsElevatorPid                   table       PID parameters for elevator adjustments in V/S mode
---     PID parameters have the following format:
---         Kp     number     proportional term coefficient (P)
---         Ki     number     integral term coefficient (I)
---         Kd     number     derivative term coefficient (D)
---         eMin   number     minimum integrated error (saturated)
---         eMax   number     maximum integrated error (saturated)
---         cMin   number     minimum command (saturated)
---         cMax   number     maximum command (saturated)
function M.setupAutopilotCustom(modelName, setup)
    if hasCalledSetup then
        log("w", "av331autopilot", "vehicle called setupAutopilotCustom after it has already been called, it has been ignored.")
        return
    end
    if not model then
        log("w", "av331autopilot", "vehicle called setupAutopilotCustom with invalid model name, it has been ignored.")
        return
    end

    hasCalledSetup = true
    if not canStillCustomSetup then
        log("w", "av331autopilot", "vehicle called setupAutopilotCustom too late, it has been ignored.")
        return
    end

    log("i", "av331autopilot", "vehicle called setupAutopilotCustom")
    model = modelName
    if setup then
        getAircraftDataHook = setup.getAircraftDataHook
        issueControls = setup.issueControls
        if setup.parameters then
            autopilotMain.applyAutopilotParameters(setup.parameters)
        end
    end
end

--- Returns the autopilot data for this vehicle.
--- @return   table   # autopilot data. contains some internal fields, the following are documented:
---      on       boolean    Autopilot engaged
---      athr     boolean    Autothrottle engaged
---      mode     string     Vertical mode: nil (none), "alt", "vs", "lvlchg"
---      spdOn    boolean    SPD mode engaged
---      hdgOn    boolean    HDG mode engaged
---      vs       number     Target vertical speed (m/s)
---      alt      number     Target altitude (m)
---      spd      number     Target speed (m/s)
---      hdg      number     Target heading (degrees)
function M.getAutopilotData()
    if not autopilot then
        return newAutopilot(model)
    end
    return autopilot
end

--- Announces the current autopilot data over an av331AutopilotChange event.
function M.announceAutopilotData()
    guihooks.trigger("av331AutopilotChange", M.getAutopilotData())
end

--- Sets an autopilot setting.
--- @param name          string      autopilot setting name
--- @param value         number      new value
function M.setAutopilotSetting(name, value)
    if not autopilot then
        M.setup()
    end
    if not autopilot then
        autopilot = newAutopilot(model)
    end
    autopilot[name] = value
    autopilotMain.configureAutopilot(autopilot, getAircraftData())
    refreshAutopilotInStream = true
end

--------------------
-- EVENT HANDLERS --
--------------------

function M.onInit()
    M.setup()
end

function M.onReset()
    local resetControls = false
    if autopilot then
        resetControls = autopilot.ok or autopilot.athr
        autopilot = newAutopilot(autopilot.model)
    end
    issueControlsModule.onReset()
    if resetControls then
        issueControls({ aileron = 0, elevator = 0, rudder = 0, throttle = 0 })
        issueControlsModule.onReset()
    end
    onAutopilotChanged()
    refreshAutopilotInStream = true
    forceSendAutopilot = true
end

local autopilotData = {}

function M.updateGFX(dt)
    if autopilot and (autopilot.on or autopilot.athr) then
        local aircraft = getAircraftData()
        local ctrl = runAutopilot(autopilot, aircraft, dt)
        local changed = false
        if ctrl.mode then
            autopilot.mode = ctrl.mode
            changed = true
        end
        if ctrl.apOff then
            autopilot.on = false
            changed = true
        end
        if ctrl.athrOff then
            autopilot.athr = false
            changed = true
        end
        if ctrl.spdOff then
            autopilot.spdOn = false
            changed = true
        end
        if ctrl.hdgOff then
            autopilot.hdgOn = false
            changed = true
        end
        if changed then
            autopilotMain.configureAutopilot(autopilot, aircraft)
            onAutopilotChanged()
        end
        issueControls(ctrl)
    end

    if forceSendAutopilot or streams.willSend("av331autopilotstream") then
        if refreshAutopilotInStream then
            local fromAutopilot = autopilot
            if not fromAutopilot then fromAutopilot = newAutopilot(nil) end

            autopilotData.on = fromAutopilot.on
            autopilotData.athr = fromAutopilot.athr
            autopilotData.mode = fromAutopilot.mode
            autopilotData.spdOn = fromAutopilot.spdOn
            autopilotData.hdgOn = fromAutopilot.hdgOn
            autopilotData.vs = fromAutopilot.vs
            autopilotData.alt = fromAutopilot.alt
            autopilotData.spd = fromAutopilot.spd
            autopilotData.hdg = fromAutopilot.hdg
            refreshAutopilotInStream = false
        end

        guihooks.send("av331autopilotstream", autopilotData)
        if forceSendAutopilot then
            forceSendAutopilot = not streams.willSend("av331autopilotstream")
        end
    end
end

return M
